/*
This sketch illustrates ideas from the course "2D Animation and Interaction"
available at www.imaginary-institute.com, and my book "Processing for
Artists and Designers," published by AK Peters.  For more of my Processing
programs, visit www.glassner.com or www.imaginary-institute.com
---
This program draws a grid of squares, along with some circles. As the circles
move, they push the squares out of their way. The squares can scale as they
move, and also rotate, creating an interesting effect. Using the keyboard,
you can choose to display a whole bunch of computer-controlled circles, or
a single circle that you control with your mouse, or both. You can change
the number of automatic circles, and change the strength with which they
push the squares away.
---
Andrew Glassner, version 2, April 2014
*/
float PushRMul = 5.0;        // how much the balls "push" the tiles out of the way
int NumAutoBalls = 5;        // number of automatic balls
int GridResolution = 18;     // number of tiles along each side of the grid
Ball[] AutoBallList;         // the computer-animated balls
Ball UserBall;               // the user-controlled ball
Tile[] TileList;             // the tiles
PVector GridUL;              // location of grid's upper-left corner
float DefaultTileR;          // default tile radius
float TileGap;               // gap between tiles
PVector MouseSelectionOffset;    // this lets us move the user's ball without jumping
boolean ActiveUserBall = true;   // include the user-controlled ball?
boolean ActiveAutoBalls = true;  // include the computer-controlled balls?
boolean RotateTiles = true;      // spin the tiles?
boolean ScaleTiles = true;       // scale the tiles?
color TileColor = color(128);
// Build the inital grid. We find the upper-left corner of the grid (GridUL),
// and make a Tile for each square in the grid. Then we create a default user-
// controlled Ball in the center, and make the list of computer-controlled
// (or "automatic") Balls.
void setup() {
  size(600, 600);
  smooth();
  float canvasSize = width * .9;
  float tileWithGap = canvasSize*1.0/GridResolution;
  DefaultTileR = (.9 * tileWithGap)/2.0;
  TileGap = tileWithGap - (2*DefaultTileR);
  float gridULx = ((width-canvasSize)/2.0) + DefaultTileR;
  float gridULy = ((height-canvasSize)/2.0) + DefaultTileR;
  GridUL = new PVector(gridULx, gridULy);
  TileList = new Tile[GridResolution * GridResolution];
  for (int sy=0; sy<GridResolution; sy++) {
    for (int sx=0; sx<GridResolution; sx++) {
      int index = (sy*GridResolution)+sx;
      float tileCx = GridUL.x + (sx * (TileGap + (2*DefaultTileR)));
      float tileCy = GridUL.y + (sy * (TileGap + (2*DefaultTileR)));
      TileList[index] = new Tile(new PVector(tileCx, tileCy), DefaultTileR);
    }
  }
  UserBall = new Ball(new PVector(width/2.0, height/2), 45, color(255, 255, 0));
  initAutoBalls();
  positionAutoBalls();
}
void draw() {
  background(255);
  if (ActiveAutoBalls) {  // if the automatic balls are enabled, draw them
    for (int i=0; i<AutoBallList.length; i++) {
      AutoBallList[i].render();
    }
  }
  if (ActiveUserBall) {  // if the user's ball is enabled, draw it
    UserBall.render();
  }
  // march through each tile and change it depending on the Balls that affect it
  for (int sy=0; sy<GridResolution; sy++) {
    for (int sx=0; sx<GridResolution; sx++) {
      int index = (sy*GridResolution)+sx;
      // get the default center, radius, rotation angle, and color for this tile
      float cx = GridUL.x + (sx * (TileGap + (2*DefaultTileR)));
      float cy = GridUL.y + (sy * (TileGap + (2*DefaultTileR)));
      float tileR = DefaultTileR;
      color tileColor = TileColor;
      float rotAngle = 0;
      /* The following is a very sneaky programming trick. I want to run the lines
      below for all the enabled balls. This might be just the user's ball, just the
      automatic balls, both, or none. I could pack it all up as a subroutine, but
      then I'd have to do some extra work to get back all the values that it
      computer (cx, cy, tileColor, and rotAngle). That could be done, but I decided
      instead to sneakify the loop. I'll start the loop at index -1, which stands for
      the user-controlled ball. I'll use the index to choose which ball I stick into
      the variable "thisBall" that gets used in the loop. And I use the booleans to
      decide whether or not to process each kind of ball. Sneaky! */
      for (int ballNumber= -1; ballNumber<AutoBallList.length; ballNumber++) {
        Ball thisBall = UserBall;
        if (ballNumber >= 0) thisBall = AutoBallList[ballNumber];
        // There's a strange interaction between the current releases of processing.js, javascript,
        // and WordPress, such that the Boolean operators && and || fail completely: if they're in 
        // the sketch, the sketch won't run! So we'll work around this by doing a few little tests
        // one at a time.
        boolean doThisBall = false;
        if (ballNumber == -1) {
          if (ActiveUserBall) {
             doThisBall = true;
          }
        }
        if (ballNumber >= 0) {
          if (ActiveAutoBalls) {
             doThisBall = true;
          }
        }
        if (doThisBall) {
          float distToBall = dist(thisBall.center.x, thisBall.center.y, cx, cy);  // dist(ball center, tile center)
          float a = distToBall/(PushRMul*thisBall.radius);    // divide this by the ball's radius scaled by push strength
          if (a < 1) {                                        // if a>1 then there will be no effect, so don't waste time
            if (ScaleTiles) tileR *= getScaleFactor(a);       // scale the radius. the factor goes down with distance
            float pushDist = getPushDistance(a, tileR, thisBall);  // how far are we pushing?
            float dx = cx-thisBall.center.x;
            float dy = cy-thisBall.center.y;
            float dlen = sqrt(sq(dx)+sq(dy));   // arrow from tile center to ball center
            cx += pushDist * dx/dlen;           // move tile center along that arrow by the scaled amount
            cy += pushDist * dy/dlen;
            tileColor = (lerpColor(thisBall.clr, tileColor, a));    // the more a tile affects us, the more we take its color
            rotAngle += map(pushDist, 0, thisBall.radius, 0, 2*PI); // and rotate as we move
          }
        }
      }
      // Now that we know that we want to go to (cx, cy) we actually only move there as if on a spring.
      // This gives us a nice, smooth motion. The spring factor was hand-tuned to look good.
      float springFactor = .95;
      TileList[index].center.x = lerp(TileList[index].center.x, cx, springFactor);
      TileList[index].center.y = lerp(TileList[index].center.y, cy, springFactor);
      TileList[index].radius = lerp(TileList[index].radius, tileR, springFactor);
      fill(tileColor);// all set. Now move to the tile center, rotate if we're doing that, and draw it
      pushMatrix();
        translate(TileList[index].center.x, TileList[index].center.y);
        if (RotateTiles) {
          rotate(rotAngle);
        }
        float r = TileList[index].radius;
        rect(-r, -r, 2*r, 2*r);
      popMatrix();
    }
  }
  positionAutoBalls();
}
// find out how far to push a tile due to a ball
float getPushDistance(float a, float tileR, Ball ball) {
  if (a>1) return(0);
  float pushDistMax = 0;
  // .707*tileR is (sqrt(2)/2)*tileR, or half the tile's diagonal length
  float pushDistMin = ball.radius + (.707*tileR);
  float d = lerp(pushDistMin, pushDistMax, a);
  return(d);
}
// blend the scale factor to look better
float getScaleFactor(float a) {
  float scaleMax = 1.0;
  float scaleMin = 0.5;
  float s = lerp(scaleMin, scaleMax, a);
  return(s);
}
// initialize the list of automatic balls
void initAutoBalls() {
  AutoBallList = new Ball[NumAutoBalls];
  for (int i=0; i<AutoBallList.length; i++) {
    float ballcx = width/2.0;
    float ballcy = height/2.0;
    float ballr = map(i, 0, AutoBallList.length-1, 50, 30);  // this can fail; see comment below
    /*
    The following line is unfortunate, but necessary. There's a bug in the current
    implementation of map() (as of Processing 2.1.1). If the first range you give map
    has identical starting and ending values, you get back either the special value
    NaN ("not a number") or the special value Infinity (also not a number!). Neither of
    these are actual floats. So if we have only one ball, the above call turns into
    map(i, 0, 0, 50, 30) and that gives us back NaN. Trying to draw an ellipse with a
    radius of NaN results in nothing. So I manually test for that special case, and set
    the radius to an arbitrary number. That's just unfortunate.
    */
    if (AutoBallList.length == 1) ballr = 40;
    colorMode(HSB);
      color ballClr = color(random(0, 255), random(128, 255), random(128, 255));
    colorMode(RGB);
    AutoBallList[i] = new Ball(new PVector(ballcx, ballcy), ballr, ballClr);
  }
}
// move the balls a little bit on every frame
void positionAutoBalls() {
  for (int i=0; i<AutoBallList.length; i++) {
    float xAngle = TWO_PI * sin((frameCount/181.0)+(i*13.371)*sin(TWO_PI*i/AutoBallList.length-1));
    float yAngle = TWO_PI * cos((frameCount/237.0)+(i*47.154)*cos(TWO_PI*i/AutoBallList.length-1));
    AutoBallList[i].center.x = (width/2.0) + ((width/4.0) * cos(xAngle));
    AutoBallList[i].center.y = (height/2.0) + ((height/4.0) * sin(yAngle));
  }
}
// respond to key presses to control the animation
void keyPressed() {
  if (key == '-') { NumAutoBalls = max(1, NumAutoBalls-1); initAutoBalls(); }   // remove an auto ball
  if (key == '+') { NumAutoBalls++;  initAutoBalls(); }                         // add another auto ball
  if (key == 'r') RotateTiles = !RotateTiles;               // toggle tile rotation
  if (key == 's') ScaleTiles = !ScaleTiles;                 // toggle tile scaling
  if (key == 'a') ActiveAutoBalls = !ActiveAutoBalls;       // toggle the automatic balls
  if (key == 'u') ActiveUserBall = !ActiveUserBall;         // toggle the user's ball
  if (key == 'b') UserBall.radius = max(10, UserBall.radius-3);  // make user's ball smaller
  if (key == 'B') UserBall.radius += 3;                     // make user's ball bigger
  if (key == 'p') PushRMul = max(1, PushRMul*.8);           // push tiles away more
  if (key == 'P') PushRMul *= 1.2;                          // push tiles away less
}
// if the mouse goes down, remember distance in x and y to the user ball so it doesn't jump
void mousePressed () {
  MouseSelectionOffset = new PVector(mouseX-UserBall.center.x, mouseY-UserBall.center.y);
}
// as the mouse is dragged, update the user's ball location
void mouseDragged() {
  if (!ActiveUserBall) return;
  UserBall.center.x = mouseX - MouseSelectionOffset.x;
  UserBall.center.y = mouseY - MouseSelectionOffset.y;
}
//------ Class Ball
// This is just a handy way to wrap up all the ball information in one place
class Ball {
  PVector center;
  float radius;
  color clr;
  Ball(PVector ballCenter, float ballRadius, color ballColor) {
    center = ballCenter.get(); // make a copy of this PVector
    radius = ballRadius;
    clr = ballColor;
  }
  void render() {
    fill(clr);
    ellipse(center.x, center.y, 2*radius, 2*radius);
  }
}
//------- Class Tile
// Another little convenience object to wrap up information about a tile
class Tile {
  PVector center;
  float radius;
  Tile(PVector tileCenter, float tileRadius) {
    center = tileCenter.get();
    radius = tileRadius;
  }
}
